@@ -4,7 +4,7 @@ use strict;
 use Carp;
 use vars qw/$VERSION $AUTOLOAD/;
 
-$VERSION = '0.98';
+$VERSION = '1.0';
 
 =head1 NAME
 
@@ -20,7 +20,7 @@ XML::Generator - Perl extension for generating XML
 
   # OR
 
-  use XML::Generator ();
+  require XML::Generator;
 
   my $X = XML::Generator->new(':pretty');
 
@@ -30,11 +30,11 @@ XML::Generator - Perl extension for generating XML
  
 Either of the above yield:
 
-   <foo>
+   <foo xmlns:qux="http://qux.com/">
      <bar baz="3">
        <bam />
      </bar>
-     <qux:bar xmlns:qux="http://qux.com/">Hey there, world</qux:bar>
+     <qux:bar>Hey there, world</qux:bar>
    </foo>
 
 =head1 DESCRIPTION
@@ -42,11 +42,12 @@ Either of the above yield:
 In general, once you have an XML::Generator object, you then simply call
 methods on that object named for each XML tag you wish to generate. 
 
-By default, C<use XML::Generator;> tries to export an C<AUTOLOAD>
-subroutine to your package, which allows you to simply call any undefined
-methods in your current package to get pieces of XML.  If you already
-have an C<AUTOLOAD> defined then XML::Generator will not override it
-unless you tell it to.  See L<"STACKABLE AUTOLOADs">.
+XML::Generator can also arrange for undefined subroutines in the caller's
+package to generate the corresponding XML, by exporting an C<AUTOLOAD>
+subroutine to your package.  Just supply an ':import' argument to your
+C<use XML::Generator;> call.  If you already have an C<AUTOLOAD> defined
+then XML::Generator can be configured to cooperate with it.  See L<"STACKABLE
+AUTOLOADs">.
 
 Say you want to generate this XML:
 
@@ -68,7 +69,7 @@ Here's a snippet of code that does the job, complete with pretty printing:
 
 The only problem with this is if you want to use a tag name that Perl's
 lexer won't understand as a method name, such as "shoe-size".  Fortunately,
-since you can always call methods as variable names, there's a simple
+since you can store the name of a method in a variable, there's a simple
 work-around:
 
    my $shoe_size = "shoe-size";
@@ -79,18 +80,45 @@ Which correctly generates:
    <shoe-size>12 1/2</shoe-size>
 
 You can use a hash ref as the first parameter if the tag should include
-atributes.  An array ref can be supplied as the first argument to indicate
+atributes.  Normally this means that the order of the attributes will be
+unpredictable, but if you have the L<Tie::IxHash> module, you can use it
+to get the order you want, like this:
+
+  use Tie::IxHash;
+  tie my %attr, 'Tie::IxHash';
+
+  %attr = (name => 'Bob', 
+	   age  => 34,
+	   job  => 'Accountant',
+    'shoe-size' => '12 1/2');
+
+  print $gen->person(\%attr);
+
+This produces
+
+  <person name="Bob" age="34" job="Accountant" shoe-size="12 1/2" />
+
+An array ref can also be supplied as the first argument to indicate
 a namespace for the element and the attributes.
 
-B<WARNING!>
+B<WARNING! ** BACKWARDS INCOMPATIBILITY **>
 
 As of version 0.94, the semantics of namespaces have changed.  Now, if there is
-one element in the array, it is considered the URI of the namespace, and the
-tag will have an xmlns="URI" attribute added automatically.  If there are two
+one element in the array, it is considered the URI of the default namespace, and
+the tag will have an xmlns="URI" attribute added automatically.  If there are two
 elements, the first should be the tag prefix to use for the namespace and the
 second element should be the URI.  In this case, the prefix will be used for
-the tag and attributes and an xmlns:PREFIX attribute will be automatically added.
-It is an error to supply more than two elements.
+the tag and an xmlns:PREFIX attribute will be automatically added.  Prior to version
+0.99, this prefix was also automatically added to each attribute name.  Now, the
+default behavior is to leave the attributes alone (although you may always explicitly
+add a prefix to an attribute name).  If the prior behavior is desired, use the
+constructor option C<qualified_attributes>.
+
+If you specify more than two elements, then each pair should correspond to a tag
+prefix and the corresponding URL.  An xmlns:PREFIX attribute will be added for each
+pair, and the prefix from the first such pair will be used as the tag's namespace.
+If you wish to specify a default namespace, use '#default' for the prefix.  If the
+default namespace is first, then the tag will use the default namespace itself.
 
 If you want to specify a namespace as well as attributes, you can make the
 second argument a hash ref.  If you do it the other way around, the array ref
@@ -98,39 +126,69 @@ will simply get stringified and included as part of the content of the tag.
 
 Here's an example to show how the attribute and namespace parameters work:
 
-   $xml = $gen->account({ type => 'checking', id => '34758'},
+   $xml = $gen->account(
 	    $gen->open(['transaction'], 2000),
 	    $gen->deposit(['transaction'], { date => '1999.04.03'}, 1500)
           );
 
 This generates:
 
-   <account type="checking" id="34578">
+   <account>
      <open xmlns="transaction">2000</open>
      <deposit xmlns="transaction" date="1999.04.03">1500</deposit>
    </account>
 
+Because default namespaces inherit, XML::Generator takes care to output
+the xmlns="URI" attribute as few times as strictly necessary.  For example,
+
+   $xml = $gen->account(
+	    $gen->open(['transaction'], 2000),
+	    $gen->deposit(['transaction'], { date => '1999.04.03'},
+	      $gen->amount(['transaction'], 1500)
+	    )
+          );
+
+This generates:
+
+   <account>
+     <open xmlns="transaction">2000</open>
+     <deposit xmlns="transaction" date="1999.04.03">
+       <amount>1500</amount>
+     </deposit>
+   </account>
+
+Notice how C<xmlns="transaction"> was left out of the C<<amount>> tag.
+
 Here is an example that uses the two-argument form of the namespace:
 
     $xml = $gen->widget(['wru' => 'http://www.widgets-r-us.com/xml/'],
                         {'id'  => 123}, $gen->contents());
 
-    <wru:widget xmlns:wru="http://www.widgets-r-us.com/xml/" wru:id="123">
+    <wru:widget xmlns:wru="http://www.widgets-r-us.com/xml/" id="123">
       <contents />
     </wru:widget>
 
-Note that the wru: prefix is inherited, so the contents tag above is
-semantically equivalent to
+Here is an example that uses multiple namespaces.  It generates the
+first example from the RDF primer (L<http://www.w3.org/TR/rdf-primer/>).
 
-      <wru:contents />
+    my $contactNS = [contact => "http://www.w3.org/2000/10/swap/pim/contact#"];
+    $xml = $gen->xml(
+	     $gen->RDF([ rdf     => "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
+		         @$contactNS ],
+		$gen->Person($contactNS, { 'rdf:about' => "http://www.w3.org/People/EM/contact#me" },
+		  $gen->fullName($contactNS, 'Eric Miller'),
+		  $gen->mailbox($contactNS, {'rdf:resource' => "mailto:em@w3.org"}),
+		  $gen->personalTitle($contactNS, 'Dr.'))));
 
-If you actually mean
-
-      <contents xmlns="" />
-
-then you have to say:
-
-    $gen->contents([undef]);
+    <?xml version="1.0" standalone="yes"?>
+    <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
+	     xmlns:contact="http://www.w3.org/2000/10/swap/pim/contact#">
+      <contact:Person rdf:about="http://www.w3.org/People/EM/contact#me">
+	<contact:fullName>Eric Miller</contact:fullName>
+	<contact:mailbox rdf:resource="mailto:em@w3.org" />
+	<contact:personalTitle>Dr.</contact:personalTitle> 
+      </Person>
+    </rdf:RDF>
 
 =head1 CONSTRUCTOR
 
@@ -163,39 +221,22 @@ Equivalent to
 	conformance => 'strict',
 	pretty      => N         # N defaults to 2
 
-=head2 :[no]import
-
-Force the exporting behavior of C<use XML::Generator;>.  Generates a
-warning when passed as an argument to C<new>.
-
-=head2 :stacked
-
-Implies :import, but if there is already an C<AUTOLOAD> defined, the
-overriding C<AUTOLOAD> will still give it a chance to run.  See L<"STACKED
-AUTOLOADs">. Generates a warning when passed as an argument to C<new>.
-
 =head2 namespace
 
-B<WARNING!>
+B<WARNING! ** BACKWARDS INCOMPATIBILITY **>
 
 As of version 0.94, this must be an array reference containing one or
 two values.  If the array contains one value, it should be a URI and will
-be the value of an 'xmlns' attribute in the top-level tag.  If there
-are two elements, the first should be the namespace tag prefix and the
-second the URI of the namespace.  This will enable behavior similar
-to the namespace behavior in previous versions; the tag prefix will be
-applied to each tag and attribute name.  In addition, an xmlns:NAME="URI"
-attribute will be added to the top-level tag.  This is consistent with
-the XML Namespaces Recommendation; the prior behavior was not.
-
-To avoid overly verbose output, XML::Generator takes care to only
-output the 'xmlns' or 'xmlns:NAME' attribute in the top-level tag, since
-namespaces are inherited.  Actually, XML::Generator outputs the 'xmlns' or
-'xmlns:NAME' attribute in the top-most tag when it is first stringified,
-which under normal usage should map to the top-level tag, but your
-mileage may vary if you are processing intermediate results.  All this
-means is that you might get more 'xmlns' or 'xmlns:NAME' attributes than
-are necessary, but the results will still be semantically correct.
+be the value of an 'xmlns' attribute in the top-level tag.  If there are
+two or more elements, the first of each pair should be the namespace
+tag prefix and the second the URI of the namespace.  This will enable
+behavior similar to the namespace behavior in previous versions; the tag
+prefix will be applied to each tag.  In addition, an xmlns:NAME="URI"
+attribute will be added to the top-level tag.  Prior to version 0.99,
+the tag prefix was also automatically added to each attribute name,
+unless overridden with an explicit prefix.  Now, the attribute names are
+left alone, but if the prior behavior is desired, use the constructor
+option C<qualified_attributes>.
 
 The value of this option is used as the global default namespace.
 For example,
@@ -212,15 +253,15 @@ would yield
 
     <HTML:html xmlns:HTML="http://www.w3.org/TR/REC-html40">
       <HTML:body>
-        <HTML:font HTML:face="Arial">Hello, there</HTML:font>
+        <HTML:font face="Arial">Hello, there</HTML:font>
       </HTML:body>
     </HTML:html>
 
-Here the same example except without all the prefixes:
+Here is the same example except without all the prefixes:
 
     my $html = XML::Generator->new(
 		 pretty    => 2,
-                 namespace => "http://www.w3.org/TR/REC-html40");
+                 namespace => ["http://www.w3.org/TR/REC-html40"]);
     print $html->html(
 	    $html->body(
 	      $html->font({ 'face' => 'Arial' },
@@ -234,29 +275,47 @@ would yield
      </body>
    </html>
 
+=head2 qualifiedAttributes, qualified_attributes
+
+Set this to a true value to emulate the attribute prefixing behavior of
+XML::Generator prior to version 0.99.  Here is an example:
+
+    my $foo = XML::Generator->new(
+                namespace => [foo => "http://foo.com/"],
+		qualifiedAttributes => 1);
+    print $foo->bar({baz => 3});
+
+yields
+
+    <foo:bar xmlns:foo="http://foo.com/" foo:baz="3" />
+
 =head2 escape
 
 The contents and the values of each attribute have any illegal XML
 characters escaped if this option is supplied.  If the value is 'always',
 then &, < and > (and " within attribute values) will be converted into
-the corresponding XML entity.  If the value is any other true value, then
-the escaping will be turned off character-by-character if the character
-in question is preceded by a backslash, or for the entire string if it
-is supplied as a scalar reference.  So, for example,
+the corresponding XML entity, although & will not be converted if it
+looks like it could be part of a valid entity (but see below).  If the
+value is 'unescaped', then the escaping will be turned off character-by-
+character if the character in question is preceded by a backslash, or for
+the entire string if it is supplied as a scalar reference.  So, for example,
 
 	use XML::Generator escape => 'always';
 
-	one('<');    # <one>&lt;</one>
-	two('\&');   # <two>\&amp;</two>
-	three(\'>'); # <three>&gt;</three> (scalar refs always allowed)
+	one('<');      # <one>&lt;</one>
+	two('\&');     # <two>\&amp;</two>
+	three(\'>');   # <three>&gt;</three> (scalar refs always allowed)
+	four('&lt;');  # <four>&lt;</four> (looks like an entity)
+	five('&#34;'); # <five>&#34;</five> (looks like an entity)
 
 but
 
-	use XML::Generator escape => 1;
+	use XML::Generator escape => 'unescaped';
 
-	one('<');    # <one>&lt;</one>
-	two('\&');   # <two>&</two>
-	three(\'>'); # <three>></three> (aiee!)
+	one('<');     # <one>&lt;</one>
+	two('\&');    # <two>&</two>
+	three(\'>');  # <three>></three> (aiee!)
+	four('&lt;'); # <four>&amp;lt;</four> (no special case for entities)
 
 By default, high-bit data will be passed through unmodified, so that
 UTF-8 data can be generated with pre-Unicode perls.  If you know that
@@ -275,13 +334,18 @@ yields
 Because XML::Generator always uses double quotes ("") around attribute
 values, it does not escape single quotes.  If you want single quotes
 inside attribute values to be escaped, use the value 'apos' along with
-'always' or 'true' for the escape option.  For example:
+'always' or 'unescaped' for the escape option.  For example:
 
     my $gen = XML::Generator->new(escape => 'always,apos');
     print $gen->foo({'bar' => "It's all good"});
 
     <foo bar="It&apos;s all good" />
 
+If you actually want & to be converted to &amp; even if it looks like it
+could be part of a valid entity, use the value 'even-entities' along with
+'always'.  Supplying 'even-entities' to the 'unescaped' option is meaningless
+as entities are already escaped with that option.
+
 =head2 pretty
 
 To have nice pretty printing of the output XML (great for config files
@@ -320,7 +384,7 @@ respectively.
 
 See L<"XML CONFORMANCE"> and L<"SPECIAL TAGS"> for more information.
 
-=head2 allowed_xml_tags
+=head2 allowedXMLTags, allowed_xml_tags
 
 If you have specified 'conformance' => 'strict' but need to use tags
 that start with 'xml', you can supply a reference to an array containing
@@ -362,6 +426,31 @@ Sets the default encoding for use in XML declarations.
 Specify the dtd.  The value should be an array reference with three
 values; the type, the name and the uri.
 
+=head1 IMPORT ARGUMENTS
+
+use XML::Generator ':option';
+
+use XML::Generator option => 'value';
+
+(Both styles may be combined)
+
+=head2 :import
+
+Cause C<use XML::Generator;> to export an C<AUTOLOAD> to your package that
+makes undefined subroutines generate XML tags corresponding to their name.
+Note that if you already have an C<AUTOLOAD> defined, it will be overwritten.
+
+=head2 :stacked
+
+Implies :import, but if there is already an C<AUTOLOAD> defined, the
+overriding C<AUTOLOAD> will still give it a chance to run.  See L<"STACKED
+AUTOLOADs">.
+
+=head2 ANYTHING ELSE
+
+If you supply any other options, :import is implied and the XML::Generator
+object that is created to generate tags will be constructed with those options.
+
 =cut
 
 package XML::Generator;
@@ -380,6 +469,7 @@ my @optionsToInit = qw(
   pretty
   version
   empty
+  qualified_attributes
 );
 
 my %tag_factory;
@@ -387,6 +477,7 @@ my %tag_factory;
 sub import {
   my $type = shift;
 
+  # check for attempt to use tag 'import'
   if (ref $type && defined $tag_factory{$type}) {
     unshift @_, $type, 'import';
     goto &{ $tag_factory{$type} };
@@ -396,20 +487,20 @@ sub import {
 
   no strict 'refs'; # Let's get serious
 
-  my $import = ( ! defined *{"${pkg}::AUTOLOAD"}{CODE} 
-		|| grep /^: (import|stacked)$/x, @_    )
-	      && ! grep /^:noimport         $/x, @_;
+  # should we import an AUTOLOAD?
+  no warnings 'once';
 
-  if ($import) {
+  if (@_) {
     my $STACKED;
 
+    # are we supposed to call their AUTOLOAD first?
     if (grep /^:stacked$/, @_) {
       $STACKED = \&{"${pkg}::AUTOLOAD"};
     }
 
     my $this = $type->new(@_);
 
-    no warnings 'redefine';
+    no warnings 'redefine'; # No, I mean SERIOUS
 
     *{"${pkg}::AUTOLOAD"} =
       sub {
@@ -422,6 +513,7 @@ sub import {
 	# The tag is whatever our sub name is.
 	my($tag) = our $AUTOLOAD =~ /.*::(.*)/;
 
+	# Special-case for xml... tags
 	if ($tag =~ /^xml/ && $this->{'conformance'} eq 'strict') {
 	  if (my $func = $this->can($tag)) {
 	    unshift @_, $this;
@@ -459,16 +551,18 @@ sub new {
 
   my %options =
     map {
-      /^:(std|standard)  $/x ? ( escape      => 'always',
-			         conformance => 'strict' )
-    : /^:strict          $/x ? ( conformance => 'strict' )
-    : /^:pretty(?:=(.+))?$/x ? ( escape      => 'always',
-			         conformance => 'strict',
-			         pretty      => ( defined $1 ? $1 : 2 ) )
-    : /^:((no)?import    |
-         stacked        )$/x ? ( do { Carp::carp("Useless use of $_")
-				        unless (caller(1))[3] =~ /::import/;
-				 () } )
+      /^:(std|standard)  $/x  ? ( escape      => 'always',
+			          conformance => 'strict' )
+    : /^:strict          $/x  ? ( conformance => 'strict' )
+    : /^:pretty(?:=(.+))?$/x  ? ( escape      => 'always',
+			          conformance => 'strict',
+			          pretty      => ( defined $1 ? $1 : 2 ) )
+    : /^:(import |
+         stacked        )$/x  ? ( do { Carp::carp("Useless use of $_")
+				         unless (caller(1))[3] =~ /::import/;
+				  () } )
+    : /^allowedXMLTags$/      ? 'allowed_xml_tags'
+    : /^qualifiedAttributes$/ ? 'qualified_attributes'
     : $_
     } @_;
 
@@ -497,7 +591,12 @@ sub new {
 	$options{'escape'} |= XML::Generator::util::ESCAPE_HIGH_BIT();
       } elsif ($1 eq 'apos') {
 	$options{'escape'} |= XML::Generator::util::ESCAPE_APOS();
+      } elsif ($1 eq 'even-entities') {
+	$options{'escape'} |= XML::Generator::util::ESCAPE_EVEN_ENTITIES();
       } elsif ($1) {
+	if ($1 ne 'unescaped') {
+	  Carp::carp "option 'escape' => '$1' deprecated; use 'escape' => 'unescaped'";
+	}
 	$options{'escape'} |= XML::Generator::util::ESCAPE_TRUE()
 			   |  XML::Generator::util::ESCAPE_GT();
       }
@@ -506,9 +605,9 @@ sub new {
     $options{'escape'} = 0;
   }
 
-  if (ref $options{'namespace'}) {
-    if (@{ $options{'namespace'} } > 2) {
-      Carp::croak "too many arguments for namespace";
+  if (ref $options{'namespace'} eq 'ARRAY') {
+    if (@{ $options{'namespace'} } > 2 && (@{ $options{'namespace'} } % 2) != 0) {
+      Carp::croak "odd number of arguments for namespace";
     }
   } elsif ($options{'namespace'}) {
     Carp::croak "namespace must be an array reference";
@@ -525,14 +624,18 @@ sub new {
 sub AUTOLOAD {
   my $this = shift;
 
-  # The tag is whatever our sub name is.
-  my ($tag) = our $AUTOLOAD =~ /.*::(.*)/;
+  # The tag is whatever our sub name is, or 'AUTOLOAD'
+  my ($tag) = defined our $AUTOLOAD ? $AUTOLOAD =~ /.*::(.*)/ : 'AUTOLOAD';
+
+  undef $AUTOLOAD; # this ensures that future attempts to use tag 'AUTOLOAD' work.
 
   unshift @_, $this, $tag;
 
   goto &{ $tag_factory{$this} };
 }
 
+# I wish there were a way to allow people to use tag 'DESTROY!'
+# hmm, maybe xmlDESTROY?
 sub DESTROY { delete $tag_factory{$_[0]} }
 
 =head1 XML CONFORMANCE
@@ -936,6 +1039,7 @@ use constant ESCAPE_HIGH_BIT => 1<<2;
 use constant ESCAPE_APOS     => 1<<3;
 use constant ESCAPE_ATTR     => 1<<4;
 use constant ESCAPE_GT       => 1<<5;
+use constant ESCAPE_EVEN_ENTITIES => 1<<6;
 
 sub parse_args {
   # this parses the args and returns a namespace and attr
@@ -951,15 +1055,15 @@ sub parse_args {
 
   # check for supplied namespace
   if (ref $args[0] eq 'ARRAY') {
-    $namespace = [ @{shift @args} ];
-    if (@$namespace > 2) {
-      croak "too many arguments for namespace";
+    $namespace = [ map { defined $_ && $_ eq '#default' ? undef : $_ } @{shift @args} ];
+    if (@$namespace > 2 && (@$namespace % 2) != 0) {
+      croak "odd number of arguments for namespace";
     }
   }
 
   # get globally-set namespace (from new)
   unless ($namespace) {
-    $namespace = [ @{ $this->{'namespace'} || [] } ];
+    $namespace = [ map { defined $_ && $_ eq '#default' ? undef : $_ } @{ $this->{'namespace'} || [] } ];
   }
 
   if (@$namespace == 1) { unshift @$namespace, undef }
@@ -967,6 +1071,10 @@ sub parse_args {
   # check for supplied attributes
   if (ref $args[0] eq 'HASH') {
     $attr = shift @args;
+    if ($this->{conformance} eq 'strict') {
+      $this->XML::Generator::util::ck_syntax($_)
+	for map split(/:/), keys %$attr;
+    }
   }
 
   return ($namespace, $attr, @args);
@@ -996,6 +1104,7 @@ sub c_tag {
              : $arg->{'pretty'}
                ? " " x $arg->{'pretty'}
                : "";
+  my $blessClass = $indent ? 'XML::Generator::pretty' : 'XML::Generator::overload';
 
   return sub {
     my $this = shift;
@@ -1051,11 +1160,13 @@ sub c_tag {
 	  if ($k =~ s/^([^:]+)://) {
 	    $this->XML::Generator::util::ck_syntax($k);
 	    $k = "$1:$k";
-	  } elsif ($prefix) {
+	  } elsif ($prefix && $this->{'qualified_attributes'}) {
 	    $this->XML::Generator::util::ck_syntax($k);
 	    $k = "$prefix$k";
+	  } else {
+	    $this->XML::Generator::util::ck_syntax($k);
 	  }
-	} else {
+	} elsif ($this->{'qualified_attributes'}) {
 	  if ($k !~ /^[^:]+:/) {
 	    $k = "$prefix$k";
 	  }
@@ -1073,21 +1184,52 @@ sub c_tag {
 	$xml .= '>';
 	if ($indent) {
 	  my $prettyend = '';
-	  
+
 	  foreach my $arg (@args) {
 	    next unless defined $arg;
-	    if ( UNIVERSAL::isa($arg, 'XML::Generator::overload') &&
-	    (! ( UNIVERSAL::isa($arg, 'XML::Generator::cdata'   ) ||
-		 UNIVERSAL::isa($arg, 'XML::Generator::pi'      )))) {
-	      $xml .= "\n$indent";
-	      $prettyend = "\n";
-	      XML::Generator::util::_fixupNS($namespace, $arg) if ref $arg->[0];
-	      $arg =~ s/\n/\n$indent/gs;
+	    if ( UNIVERSAL::isa($arg, 'XML::Generator::cdata'   ) ) {
+	      my $copy = $xml;
+	      push @xml, $copy, $arg;
+	      $xml = '';
+	    } else {
+	      if ( UNIVERSAL::isa($arg, 'XML::Generator::overload') &&
+	           ! UNIVERSAL::isa($arg, 'XML::Generator::pi') ) {
+		$xml .= "\n$indent";
+		$prettyend = "\n";
+		XML::Generator::util::_fixupNS($namespace, $arg) if ref $arg->[0];
+
+		my @cdata;
+		for my $i (0..$#$arg) {
+		  if (UNIVERSAL::isa($arg->[$i], 'XML::Generator::cdata')) {
+		    push @cdata, $arg->[$i];
+		    $arg->[$i] = "\001";
+		  }
+		}
+
+		$arg =~ s/\n/\n$indent/gs;
+
+		if (@cdata) {
+		  my @pieces = split "\001", $arg;
+
+		  my $copy = $xml;
+		  push @xml, $copy;
+		  $xml = '';
+		  $arg = '';
+
+		  for my $i (0..$#pieces) {
+		    if (defined $cdata[$i]) {
+		      push @xml, $pieces[$i], $cdata[$i];
+		    } else {
+		      push @xml, $pieces[$i];
+		    }
+		  }
+		}
+	      }
+	      $xml .= "$arg";
 	    }
-	    $xml .= "$arg";
 	  }
 	  $xml .= $prettyend;
-	  @xml = ($xml, "</$prefix$tag>");
+	  push @xml, ($xml, "</$prefix$tag>");
 	} else {
 	  @xml = $xml;
 	  foreach my $arg (grep defined, @args) {
@@ -1111,19 +1253,28 @@ sub c_tag {
 
     unshift @xml, $namespace if $namespace;
 
-    return XML::Generator::overload->new(\@xml);
+    return $blessClass->new(\@xml);
   };
 }
 
 sub _fixupNS {
   # remove namespaces
-  # if prefix and uri match one we have, remove them from child
-  # if prefix does not match one we have, remove it and uri from child
-  #   and add them to us
+  # if prefix
+  #    if prefix and uri match one we have, remove them from child
+  #    if prefix does not match one we have, remove it and uri
+  #      from child and add them to us
+  # no prefix
+  #    if we have an explicit default namespace and the child has the
+  #      same one, remove it from the child
+  #    if we have an explicit default namespace and the child has a
+  #      different one, leave it alone
+  #    if we have an explicit default namespace and the child has none,
+  #      add an empty default namespace to child
   my($namespace, $o) = @_;
   my @n = @{$o->[0]};
+  my $sawDefault = 0;
   for (my $i = 0; $i < $#n; $i+=2) {
-    if (defined $n[$i]) {
+    if (defined $n[$i]) { # namespace w/ prefix
       my $flag = 0;
       for (my $j = 0; $j < $#$namespace; $j+=2) {
 	next unless defined $namespace->[$j];
@@ -1142,21 +1293,26 @@ sub _fixupNS {
 	splice @n, $i, 2;
 	$i-=2;
       }
-    } elsif (defined $n[$i+1]) {
-      my $flag = 0;
+    } elsif (defined $n[$i+1]) { # default namespace
+      $sawDefault = 1;
       for (my $j = 0; $j < $#$namespace; $j+=2) {
 	next if defined $namespace->[$j];
-	$flag = 1;
         if ($namespace->[$j+1] eq $n[$i+1]) {
 	  splice @n, $i, 2;
 	  $i-=2;
 	}
       }
-      unless ($flag) {
-	push @$namespace, splice @n, $i, 2;
-      }
     }
   }
+
+  # check to see if we need to add explicit default namespace of "" to child
+  if (! @{ $o->[0] } &&
+      ! $sawDefault &&
+      grep { defined $namespace->[$_ * 2 + 1] &&
+           ! defined $namespace->[$_ * 2    ] } 0..($#$namespace/2)) {
+    push @n, undef, "";
+  }
+
   if (@n) {
     $o->[0] = [@n];
   } else {
@@ -1180,8 +1336,12 @@ sub escape {
   return unless defined $_[0];
 
   my $f = $_[1];
-  if ($_[1] & ESCAPE_ALWAYS) {
-    $_[0] =~ s/&(?!(#[0-9]+|#x[0-9a-fA-F]+|\w+);)/&amp;/g;
+  if ($f & ESCAPE_ALWAYS) {
+    if ($f & ESCAPE_EVEN_ENTITIES) {
+      $_[0] =~ s/&/&amp;/g;
+    } else {
+      $_[0] =~ s/&(?!(?:#[0-9]+|#x[0-9a-fA-F]+|\w+);)/&amp;/g;
+    }
 
     $_[0] =~ s/</&lt;/g;
     $_[0] =~ s/>/&gt;/g   if $f & ESCAPE_GT;
@@ -1272,10 +1432,11 @@ sub get_dtd {
 
 package XML::Generator::overload;
 
-use overload '""'   => \&stringify,
-             '0+'   => \&stringify,
-             'bool' => \&stringify,
-             'eq'   => sub { stringify($_[0]) eq stringify($_[1]) };
+use overload '""'   => sub { $_[0]->stringify },
+             '0+'   => sub { $_[0]->stringify },
+             'bool' => sub { $_[0]->stringify },
+             'eq'   => sub { (ref $_[0] ? $_[0]->stringify : $_[0]) eq
+			     (ref $_[1] ? $_[1]->stringify : $_[1])};
 
 sub new {
   my($class, $xml) = @_;
@@ -1286,7 +1447,7 @@ sub stringify {
   return $_[0] unless UNIVERSAL::isa($_[0], 'XML::Generator::overload');
   if (ref($_[0]->[0])) { # namespace
     my $n = shift @{$_[0]};
-    for (my $i = 0; $i < @$n; $i+=2) {
+    for (my $i = ($#$n - 1); $i >= 0; $i-=2) {
       my($prefix, $uri) = @$n[$i,$i+1];
       XML::Generator::util::escape($uri, XML::Generator::util::ESCAPE_ATTR  |
                                          XML::Generator::util::ESCAPE_ALWAYS|
@@ -1299,11 +1460,28 @@ sub stringify {
       }
     }
   }
+
   join $, || "", @{$_[0]}
 }
 
 sub DESTROY { }
 
+package XML::Generator::pretty;
+
+use base 'XML::Generator::overload';
+
+sub stringify {
+  my $this = shift;
+  my $string = $this->SUPER::stringify();
+
+  $string =~ s{^((\s*<(?:\w+:)?\w[-.\w]* )[^ "]+"[^"]+")( .{40,})}
+	      {  my($a,$b,$c) = ($1, $2, $3);
+		$c =~ s{ ((?:\w+:)?\w+="[^\"]+")}{"\n" . (' 'x(length $b)) . $1}ge;
+		"$a$c" }gem;
+
+  return $string;
+}
+
 package XML::Generator::final;
 
 use base 'XML::Generator::overload';
